# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, unicode_literals, division, print_function
)

import sys
from functools import wraps

from tornado.gen import is_future


def fail_to(future):
    """A decorator for function callbacks to catch uncaught non-async
    exceptions and forward them to the given future.

    The primary use for this is to catch exceptions in async callbacks and
    propagate them to futures. For example, consider,

    .. code-block:: python

        answer = Future()

        def on_done(future):
            foo = bar()
            answer.set_result(foo)

        some_async_operation().add_done_callback(on_done)

    If ``bar()`` fails, ``answer`` will never get filled with an exception or
    a result. Now if we change ``on_done`` to,

    .. code-block:: python

        @fail_to(answer)
        def on_done(future):
            foo = bar()
            answer.set_result(foo)

    Uncaught exceptions in ``on_done`` will be caught and propagated to
    ``answer``. Note that ``on_done`` will return None if an exception was
    caught.

    :param answer:
        Future to which the result will be written.
    """
    assert is_future(future), 'you forgot to pass a future'

    def decorator(f):

        @wraps(f)
        def new_f(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception:
                future.set_exc_info(sys.exc_info())

        return new_f

    return decorator



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import os
import os.path
from itertools import chain
from collections import deque
from collections import namedtuple

from . import yaml
from .exceptions import RequestNotFoundError
from .exceptions import UnsupportedVersionError
from .record_modes import RecordMode
from . import proxy

__all__ = ['Cassette', 'DEFAULT_MATCHERS']


# Version of the storage format.
VERSION = 1


class Interaction(namedtuple('Interaction', 'request response')):
    """An interaction is a request-response pair."""

    def to_primitive(self):
        return {
            'request': self.request.to_primitive(),
            'response': self.response.to_primitive(),
        }

    @classmethod
    def to_native(cls, data):
        return cls(
            request=proxy.Request.from_primitive(data['request']),
            response=proxy.Response.from_primitive(data['response']),
        )


def attrmatcher(name):
    "A matcher that matches if the given attribute is equal for both values."

    def matcher(left, right):
        return getattr(left, name) == getattr(right, name)

    return matcher

# A dictionary from matcher name to matcher.
#
# A matcher is a function that takes two arguments and returns true if they
# match.
#
# The dictionary contains all known matchers.
_MATCHERS = {
    n: attrmatcher(n) for n in (
        'serviceName', 'hostPort', 'endpoint', 'headers', 'body', 'argScheme',
        'transportHeaders',
    )
}


DEFAULT_MATCHERS = (
    'serviceName', 'endpoint', 'headers', 'body', 'argScheme',
)


class Cassette(object):
    """Represents a series of recorded interactions."""

    def __init__(self, path, record_mode=None, matchers=None):
        """Initialize a new cassette.

        :param path:
            File path at which this cassette will be stored.
        :param record_mode:
            One of 'once', 'none', 'all', 'new_episodes'. See
            :py:class:`tchannel.testing.vcr.RecordMode`.
        :param matchers:
            If specified, this is a collection of matcher names. These
            specify which attributes on two requests should match for them to
            be considered equal.
        """
        # TODO move documentation around
        record_mode = record_mode or RecordMode.ONCE
        if isinstance(record_mode, basestring):
            record_mode = RecordMode.from_name(record_mode)

        self.path = path

        # Whether the cassette was loaded from an existing YAML. If False,
        # this was a new cassette and the YAML file did not exist.
        self.existed = False

        if matchers is None:
            matchers = DEFAULT_MATCHERS

        self._matchers = []
        for m in matchers:
            try:
                self._matchers.append(_MATCHERS[m])
            except KeyError:
                raise KeyError('%s is not a known matcher' % m)

        self._record_mode = record_mode
        self._available = deque()
        self._played = deque()
        self._recorded = deque()

        self._load()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.save()

    @property
    def record_mode(self):
        """The record mode being used for this cassette."""
        return self._record_mode.name

    @property
    def write_protected(self):
        return not self._record_mode.can_record(self)

    @property
    def play_count(self):
        """Number of responses that have been replayed."""
        return len(self._played)

    @property
    def data(self):
        """Get all known data for this cassette."""

        return deque(
            chain(self._played, self._available, self._recorded)
        )

    def _match(self, left, right):
        return all(m(left, right) for m in self._matchers)

    def _find(self, request, haystack=None):
        "A generator that returns yields matching interactions for a request"

        haystack = haystack or self._available

        for interaction in haystack:
            if self._match(request, interaction.request):
                yield interaction

    def _load(self):
        try:
            with open(self.path, 'rb') as f:
                data = f.read()
                self.existed = True
        except IOError:
            return  # nothing to read

        data = yaml.load(data)
        if not (data and 'interactions' in data):
            return  # file was probably empty

        if int(data['version']) != VERSION:
            raise UnsupportedVersionError(
                'Cassette at "%s" is an unsupported version of the '
                'format: version %s' % (self.path, str(data['version']))
            )

        self._available = [
            Interaction.to_native(i) for i in data['interactions']
        ]

    def save(self):
        if not self._recorded:
            return

        # Order:
        # - things that were played in the order that were played
        # - things that haven't been played yet -- assuming the record mode
        #   allows it.
        # - things that were recorded in this session
        interactions = deque(self._played)
        if self._record_mode.save_unplayed:
            interactions.extend(self._available)
        interactions.extend(self._recorded)

        data = yaml.dump(
            {
                'interactions': [i.to_primitive() for i in interactions],
                'version': VERSION,
            }
        )

        cassette_dir = os.path.dirname(self.path)
        if not os.path.isdir(cassette_dir):
            os.makedirs(cassette_dir)

        with open(self.path, 'wb') as f:
            f.write(data)

        self._played = deque()
        self._available = interactions
        self._recorded = deque()

    def can_replay(self, request):
        if not self._record_mode.replayable:
            return False
        try:
            next(self._find(request))
        except StopIteration:
            return False
        else:
            return True

    def replay(self, request):
        assert self._record_mode.replayable, (
            'The record mode for this cassette prevents it from replaying '
            'requests'
        )

        for interaction in self._find(request):
            # TODO request matchers
            self._available.remove(interaction)
            self._played.append(interaction)
            return interaction.response

        raise RequestNotFoundError(
            'Could not find a recorded response for %s' % repr(request)
        )

    def record(self, request, response):
        assert not self.write_protected, (
            'The record mode for this cassette prevents it from recording '
            'new requests'
        )
        self._recorded.append(Interaction(request, response))



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module implements tombstones for TChannel connections.

A tombstone is a marker for a request that timed out. Often, shortly after a
request times out, we receive a message for it anyway. This is because the
server's timeout is often skewed. So if we immediately forget about the
request on timeout, we'll see an "unconusmed message" error when we receive
these zombie messages. To avoid this, we leave a tombstone behind for the
request that timed out so that we know it is safe to ignore the zombie
messages.

Tombstones are destroyed automatically after a fixed duration.
"""

from __future__ import (
    absolute_import, unicode_literals, print_function, division
)

from tornado.ioloop import IOLoop


# Default offset of time (in seconds) on top of the original request TTL for
# which the tombstone will be active.
DEFAULT_TTL_OFFSET_SECS = 0.5

# Default maximum amount of time (in seconsd) for which a tombstone can be
# active.
DEFAULT_MAX_TTL_SECS = 5


class Cemetery(object):
    """Cemetery is a collection of tombstones.

    A tombstone is just a unique identifier for requests that timed out
    locally.

    :param ttl_offset_secs:
        Amount of time (in seconds) on top of the original request's timeout
        for which the tombstone for that request will exist.
    :param max_ttl_secs:
        Maximum amount of time (in seconds) for which a tombstone for a
        request can exist.
    """

    __slots__ = ('_tombstones', 'io_loop', 'ttl_offset_secs', 'max_ttl_secs')

    def __init__(self, ttl_offset_secs=None, max_ttl_secs=None):
        if ttl_offset_secs is None:
            ttl_offset_secs = DEFAULT_TTL_OFFSET_SECS

        if max_ttl_secs is None:
            max_ttl_secs = DEFAULT_MAX_TTL_SECS

        self._tombstones = {}
        self.ttl_offset_secs = ttl_offset_secs
        self.max_ttl_secs = max_ttl_secs

    def __contains__(self, id):
        """Check if the request with the given id is known to have timed
        out."""
        return id in self._tombstones

    def add(self, id, ttl_secs):
        """Adds a new request to the Cemetery that is known to have timed out.

        The request will be forgotten after ``ttl_secs + ttl_offset_secs``
        seconds.

        :param id:
            ID of the request
        :param ttl_secs:
            TTL of the request (in seconds)
        """
        ttl_secs = min(ttl_secs + self.ttl_offset_secs, self.max_ttl_secs)
        self._tombstones[id] = IOLoop.current().call_later(
            ttl_secs, self.forget, id,
        )

    def forget(self, id):
        """Forget about a specific request."""
        self._tombstones.pop(id, None)

    def clear(self):
        """Forget about all requests."""
        io_loop = IOLoop.current()
        while self._tombstones:
            _, req_timeout = self._tombstones.popitem()
            io_loop.remove_timeout(req_timeout)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from . import common
from .. import rw
from .call_continue import CallContinueMessage
from .types import Types


class CallRequestContinueMessage(CallContinueMessage):
    """Represent a continuation of a call request (across multiple frames)."""
    message_type = Types.CALL_REQ_CONTINUE

    def __init__(
        self,
        flags=0,
        checksum=None,
        args=None,
        id=0,
    ):
        super(CallRequestContinueMessage, self).__init__(
            flags, checksum, args, id)

    def fragment(self, space_left):
        fragment_msg = CallRequestContinueMessage(
            flags=self.flags,
            checksum=self.checksum,
        )
        return super(CallRequestContinueMessage, self).\
            fragment(space_left, fragment_msg)


call_req_c_rw = rw.instance(
    CallRequestContinueMessage,
    ("flags", rw.number(1)),    # flags:1
    ("checksum", common.checksum_rw),   # csumtype:1 (csum:4){0, 1}
    ("args", rw.args(rw.number(2))),    # [arg1~2, arg2~2, arg3~2]
)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from tchannel.thrift.reflection import get_service_methods
from tchannel.thrift.client import client_for as async_client_for


def client_for(service, service_module, thrift_service_name=None):
    """Build a synchronous client class for the given Thrift service.

    The generated class accepts a TChannelSyncClient and an optional
    hostport as initialization arguments.

    Given ``CommentService`` defined in ``comment.thrift`` and registered
    with Hyperbahn under the name "comment", here's how this might be used:

    .. code-block:: python

        from tchannel.sync import TChannelSyncClient
        from tchannel.sync.thrift import client_for

        from comment import CommentService

        CommentServiceClient = client_for('comment', CommentService)

        tchannel_sync = TChannelSyncClient('my-service')
        comment_client = CommentServiceClient(tchannel_sync)

        future = comment_client.postComment(
            articleId,
            CommentService.Comment("hi")
        )
        result = future.result()

    :param service:
        Name of the Hyperbahn service being called.
    :param service_module:
        The Thrift-generated module for that service. This usually has
        the same name as definied for the service in the IDL.
    :param thrift_service_name:
        If the Thrift service has a different name than its module, use
        this parameter to specify it.
    :returns:
        An Thrift-like class, ready to be instantiated and used
        with TChannelSyncClient.
    """
    assert service_module, 'service_module is required'

    service = service or ''  # may be blank for non-hyperbahn use cases

    if not thrift_service_name:
        thrift_service_name = service_module.__name__.rsplit('.', 1)[-1]

    method_names = get_service_methods(service_module.Iface)

    def init(
        self,
        tchannel,
        hostport=None,
        trace=False,
        protocol_headers=None,
    ):
        self.async_thrift = self.__async_client_class__(
            tchannel=tchannel,
            hostport=hostport,
            trace=trace,
            protocol_headers=protocol_headers,
        )
        self.threadloop = tchannel._threadloop

    init.__name__ = '__init__'

    methods = {
        '__init__': init,
        '__async_client_class__': async_client_for(
            service=service,
            service_module=service_module,
            thrift_service_name=thrift_service_name,
        )
    }

    methods.update({
        method_name: generate_method(method_name)
        for method_name in method_names
    })

    return type(thrift_service_name + 'Client', (object,), methods)


def generate_method(method_name):
    """Generate a method for a given Thrift service.

    Uses the provided TChannelSyncClient's threadloop in order
    to convert RPC calls to concurrent.futures

    :param method_name: Method being called.
    :return: A method that invokes the RPC using TChannelSyncClient
    """

    def call(self, *args, **kwargs):
        """Forward RPC call to TChannelSyncClient

        :return concurrent.futures.Future:
        """
        if not self.threadloop.is_ready():
            self.threadloop.start()

        return self.threadloop.submit(
            getattr(self.async_thrift, method_name), *args, **kwargs
        )

    return call
